/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.camel.component.sed;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;


/**
 *  Container class for a set of SED-like expressions. The class supports reading
 *  the expressions from a configuration file, or manual specification of one or
 *  more expressions.  For now, expressions are limited to the following general
 *  form:
 *
 *  's/${findPattern}/${replacePattern}/${optionalFlags}'
 *
 *  Specifying expressions in a simple external file rather than in-line Java code 
 *  follows the general guidance given in the book 'Java Regular Expressions'.
 *  The basic idea is to remove the need for double-escaping the patterns - once
 *  for the Java compiler and once for the regular expression engine (can be
 *  confusing)
 *
 */
public class SedExpressionSet {

  private static final transient Log log =
    LogFactory.getLog( SedExpressionSet.class );

  Pattern       _ruleParsePattern;
  List<SedExpression> _ruleList;
  File _file = null;
  

  /**
   *  Default constructor
   */
  public SedExpressionSet()
  {
    //
    // Define a pattern that breaks up the SED find/replace expression lines into
    // their components. Delimeter is '/', unless preceded by an escaping 
    // '\'. This pattern uses lookbehind syntax to check for preceding '\'. 
    // Note that the pattern has to specify 4 '\' characters since the Java compiler 
    // unescapes the 4 down to 2, and in the pattern itself, a '\' character
    // match must be escaped by a preceding '\'. 
    //
    _ruleParsePattern = Pattern.compile("(?<!\\\\)/",Pattern.MULTILINE);

    _ruleList = new ArrayList<SedExpression>();
  }


  /**
   * Convenience constructor that sets up a SED rule set with the single
   * specified rule.
   *
   * @param sedExpression  Single initial SED rule for the rule set
   *
   * @throws  PatternSyntaxException if the rule isn't properly formed
   */
  public SedExpressionSet( String sedExpression )
    throws PatternSyntaxException {

    this();
    addExpression( sedExpression );
  }

  /**
   *  Loads Sed rule set from a file. @See load(FileInputStream inStream)
   *
   * @param file The name of the file to load
   *
   * @throws IOException if there's an IO problem
   * @throws  PatternSyntaxException if the File format isn't properly
   * formed
   */
  public SedExpressionSet( File file )
    throws IOException, PatternSyntaxException {

    this();
    _file = file;  // Save for error message annotations
    processStream( new FileInputStream( _file ) );
  }


  public SedExpressionSet( InputStream inStream )
    throws IOException, PatternSyntaxException {

    this();
    processStream( inStream );
  }


  /**
   * Process an input stream containing one or more SED rules
   *
   * @param inStream  File input stream
   *
   * @throws IOException if there's an IO problem
   * @throws PatternSyntaxException if the patterns in the file aren't properly
   * formed (fail the regex compiler stage)
   *
   */
  public void processStream( InputStream inStream )
    throws IOException, PatternSyntaxException {

    // load the contents of the file

    BufferedReader reader = new BufferedReader( new InputStreamReader( inStream ) );
    
    String line;
    int lineNum = 1;  // Track line number for error reporting

    while( (line = reader.readLine()) != null ) {

      // If the line is null or a comment, ignore it
      if( line != null &&  !"".equals(line.trim()) && !line.startsWith("#") )
      {
        log.debug( "Processing line: " + line );
        try {
          addExpression( line );
        }
        catch( PatternSyntaxException pse ) {
          log.error( "SED rule syntax error " + errorLocationString( lineNum ) );
          // Rethrow regex library-generated exception
          throw pse;
        }
      }

      lineNum++;
    }

    reader.close();
  }


  /**
   * Add a single SED expression to the set. Lines are rules of the general form: 
   *
   *  's/${findPattern}/${replacePattern}/${optionalFlags}'
   *
   *  Parse each line, break it up into its components, compile
   *  the patterns, and store the information in a rule list
   *
   * @param line   Expression line
   */
  public void addExpression( String line ) 
    throws PatternSyntaxException
  {
    String[] fields = _ruleParsePattern.split( line );

    //
    // For now a relatively fixed rule syntax is supported ( 's/$patt/$patt/[optionalFlags]' )
    //
    if( (fields.length < 3) || (fields.length > 4) )
    {
      throw new PatternSyntaxException( "Expected 3 or 4 fields in SED search/replace rule, found " + 
                                        fields.length, 
                                        line, 0 );
    }

    //
    // Un-escape '/' characters in find/replace patterns that were escaped to 
    // avoid collision with SED sytax's '/' field delimiter 
    //
    String findPatternStr = fields[1].replaceAll("\\\\/", "/" );
    String replacePatternStr = fields[2].replaceAll("\\\\/", "/" );
        
    // 
    // Sed/Perl form of regular expressions supports the matching group
    // variable syntax of '\N', where N is a number between 0-9. Java
    // uses the '$N' nomenclature instead. Convert the pattern here for
    // the replace pattern string. This rule reads:
    //
    //  "Find occurrences of '\N', where N is a digit between [0-9], and
    //   store the value of N in the variable $1. Replace the '\N' with
    //   that with '$N' recovering the actual value of N from the
    //   variable $1"
    //
    replacePatternStr = replacePatternStr.replaceAll("\\\\([0-9])", "\\$$1" );

    String options;
    if( fields.length > 3 )  // Trailing options field present?
      options = fields[3];
    else
      options = "";

    //
    // If the trailing options field contains an 'i'  (e.g 's/searchPattern/replacePattern/gi')
    // then match is case insensitive. Note that the 'i' and 'g' are independent options
    //
    Pattern findPattern;
    if( options.indexOf('i') >= 0 )
      findPattern = Pattern.compile( findPatternStr, Pattern.CASE_INSENSITIVE );
    else
      findPattern = Pattern.compile( findPatternStr );

    SedExpression rule = new SedExpression( fields[0], findPattern, replacePatternStr, options );
    _ruleList.add( rule );
    
  }
  

  public List<SedExpression> getExpressionList() 
  {
    return _ruleList;
  }
  
  private String errorLocationString( int lineNum ) 
  {
    return "(" + getInputStreamName() + ":" + lineNum + ")";
  }

  private String getInputStreamName() 
  {
    if( _file == null )
    {
      return "Unknown input stream";
    }
    else
    {
      String path;
      try {
        path = _file.getCanonicalPath();
      }
      catch( IOException e ) {
        path = _file.getAbsolutePath();
      }
      
      return path;
    }
  }

}
 
