/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.camel.component.sed;

import org.apache.camel.Exchange;
import org.apache.camel.Message;
import org.apache.camel.Processor;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.regex.Matcher;
import java.util.regex.PatternSyntaxException;



/**
 *  Camel processor that transforms messages based on regEx
 *  regular expressions, specified using  SED/Perl-like syntax, i.e.:
 *
 *     's/version1.4/version1.3/g;'
 *  
 *  Multiple expressions can be specified in a configuration file for
 *  simple one-pass editing.
 *
 */
public class SedProcessor implements Processor
{
  private static final transient Log log = 
    LogFactory.getLog( SedProcessor.class );

  SedExpressionSet _expressionSet = null;
  

  /**
   *  Default constructor. Useful in Spring context.
   */
  public SedProcessor() {
  }

  /**
   *  Convenience constructor that initializes the processor with a single
   *  SED expression.
   *
   *  @param sedExpression   Single constructor
   *
   *  @throws PatternSyntaxException if there is a syntax error in the expression.
   *
   */
  public SedProcessor( String sedExpression )
    throws PatternSyntaxException {

    _expressionSet = new SedExpressionSet( sedExpression );
  }

  /**
   *  File variant of constructor.
   *
   *  @param file   Configuration file
   *
   *  @throws IOException if file doesn't exist
   *  @throws PatternSyntaxException if there is a syntax error in any of the
   *  expressions in the configuration file.
   */
  public SedProcessor( File file )
    throws IOException, PatternSyntaxException
  {
    _expressionSet = new SedExpressionSet( file );
  }
  

  /**
   *  Generic input stream constructor
   *
   *  @param inStream   Input stream for configuration file. 
   *
   */
  public SedProcessor( InputStream inStream )
    throws IOException, PatternSyntaxException
  {
    _expressionSet = new SedExpressionSet( inStream );
  }


  /**
   *  Set the name of the configuration file, and load the file.
   *  This method allows for a SedProcessor to be easily configured
   *  in Spring, i.e.:
   *
   *  <pre>
   *    <bean id="simpleServiceVersionProcessor"
   *          class="org.apache.camel.component.sed.SedProcessor">
   *      <property name="filename" value="./src/main/resources/serviceVersionRegexConfig"/>
   *    </bean>
   *  </pre>
   *
   *  @param filename   Configuration filename
   *
   *
   *  This version is for specifying a file in the filesystem. See setConfigResource()
   *  for setting to a stream in the classpath
   *
   */
  public void setConfigFilename( String filename )
    throws IOException, PatternSyntaxException
  {
    _expressionSet = new SedExpressionSet( new File(filename) );
  }

  /**
   *  Set the name of the configuration resource. This is a variant 
   *  of setConfigFilename() that allows file to be located via the 
   *  classloader.
   *  in Spring, i.e.:
   *
   *  <pre>
   *    <bean id="simpleServiceVersionProcessor"
   *          class="org.apache.camel.component.sed.SedProcessor">
   *      <property name="resource" value="serviceVersionRegexConfig"/>
   *    </bean>
   *  </pre>
   *
   *  @param resource   Configuration resource
   *
   */
  public void setConfigResource( String resource )
    throws IOException, PatternSyntaxException
  {
    InputStream inStream = getClass().getResourceAsStream( resource );
    _expressionSet = new SedExpressionSet( inStream );
  }

  
  /**
   *  Process the message, applying the set of all the regex expressions to 
   *  the message in turn
   *
   *  Note: The SED processing only works on the message body for now. Headers
   *  and/or attachments are simply copied through.
   *
   *  @param exchange   Message exchange object
   */
  public void process( Exchange exchange ) throws Exception 
  {
    if( _expressionSet == null )
      throw new IOException("No config file loaded for SedProcessor instance");


    Message inMsg = exchange.getIn();
    String inBody = inMsg.getBody( String.class );
    
    //
    // If this message is output-capable, copy the unmodified inMsg to the
    // output message. This ensures that msg headers and attachements pass 
    // through the processor 
    //
    if( exchange.getPattern().isOutCapable() )
      exchange.getOut().copyFrom( inMsg );

    if( log.isDebugEnabled() )
      log.debug("Body before regex: " + inBody );
    
    String tmpBody = inBody;

    for( SedExpression expression : _expressionSet.getExpressionList() )
    {
      Matcher matcher = expression.getFindPattern().matcher( tmpBody );
      if( expression.getReplaceAllFlag() )
        tmpBody = matcher.replaceAll( expression.getReplacePattern() );
      else
        tmpBody = matcher.replaceFirst( expression.getReplacePattern() );
    }

    if( log.isDebugEnabled() )
      log.debug( "Body after regex: " + tmpBody );
    
    if( exchange.getPattern().isOutCapable() )
    {
      exchange.getOut().setBody( tmpBody );
    }
    else
    {
      exchange.getIn().setBody( tmpBody );
    }

  }

}



