/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.camel.component.sed;

import static org.testng.AssertJUnit.*;
import org.testng.annotations.AfterTest;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.Test;

import org.apache.camel.Processor;
import org.apache.camel.Exchange;
import org.apache.camel.ExchangePattern;
import org.apache.camel.Message;
import org.apache.camel.CamelContext;

import org.apache.camel.impl.DefaultCamelContext;
import org.apache.camel.impl.DefaultExchange;


import java.util.Map;
import java.util.List;
import java.io.File;
import java.io.IOException;
import java.util.regex.*;


/**
 *  Regex processor test. Load configuration file and run processor on 
 *  a variety of messages
 */
public class SedProcessorTest {

  CamelContext _camelContext;
  Exchange     _exchange;
  Message      _inMsg;
  Message      _outMsg;


  @BeforeTest
  public void setUp()
    throws IOException {

    _camelContext = new DefaultCamelContext();
    _exchange = new DefaultExchange( _camelContext );
    // Default Exchange pattern is InOnly - use InOut here 
    _exchange.setPattern( ExchangePattern.InOut );

    _inMsg = _exchange.getIn();
    _outMsg = _exchange.getOut();
  }

  @AfterTest
  public void tearDown() {
    // Add your code here
  }


  /** 
   *  Test of editing a namespace in an XML message from:
   *
   *   "http://ll.mit.edu/fuse/examples/simpleService/1.1"
   *
   *  to:
   *
   *   "http://ll.mit.edu/fuse/examples/notSoSimpleService/1.1"
   *
   *  This version creates a processor instance with a single SED 
   *  rule using the convenience constructor that takes a single string
   *  argument
   */
  @Test
  public void modifyXmlNamespaceSimpleTest()
    throws IOException, PatternSyntaxException, Exception {

    //
    // Create the process and initialize with a single SED expression 
    // (most commonly, expressions will be stored in a file - see subsequent
    // tests)
    //
    SedProcessor processor = 
      new SedProcessor("s/simpleService/notSoSimpleService/g" );

    //
    // Set up a couple of test headers in the input message. Currently the
    // SedProcessor doesn't do any filtering on message headers (syntax TBD),
    // but it is set up to simply copy them to the output message (for InOut exchanges)
    //
    _inMsg.setHeader("headerName1", "headerValue1" );
    _inMsg.setHeader("headerName2", "headerValue2" );
    

    // Set up the test body
    _inMsg.setBody("http://ll.mit.edu/fuse/examples/simpleService/1.1");
    

    // Process the message using the SED expression
    processor.process( _exchange );

    // Verify that headers make it through 
    Map<String,Object> hdrMap = _outMsg.getHeaders();

    assertTrue( hdrMap.size() == 2 );
    assertTrue( _outMsg.getHeader("headerName1") != null );
    assertTrue( _outMsg.getHeader("headerName2") != null );
    

    String outBody = _outMsg.getBody( String.class );
    //System.out.println("Message outMsg body: " + outBody );

    assertTrue( outBody.equals("http://ll.mit.edu/fuse/examples/notSoSimpleService/1.1") );

  }

  /** 
   *  Test of editing a namespace version suffix in an XML message from:
   *
   *   "http://ll.mit.edu/fuse/examples/simpleService/1.X"
   *
   *  to:
   *
   *   "http://ll.mit.edu/fuse/examples/simpleService/1.0"
   *
   *  This version initializes a SedProcessor instance using a file containing
   *  multiple SED expressions
   */
  @Test
  public void modifyXmlNamespaceVersionTest()
    throws IOException, PatternSyntaxException, Exception {

    SedProcessor processor = 
      new SedProcessor( new File("./src/test/resources/modifyXmlNamespaceVersion") );

    // Simulate a version 1.1 Xml namespace
    _inMsg.setBody("xmlns=\"http://ll.mit.edu/fuse/examples/simpleService/1.1\"");

    processor.process( _exchange );

    String outBody = _outMsg.getBody( String.class );

    assertTrue( outBody.equals("xmlns=\"http://ll.mit.edu/fuse/examples/simpleService/1.0\"") );


    // Simulate a version 1.2 Xml namespace
    _inMsg.setBody("xmlns=\"http://ll.mit.edu/fuse/examples/simpleService/1.2\"");

    processor.process( _exchange );

    outBody = _outMsg.getBody( String.class );

    assertTrue( outBody.equals("xmlns=\"http://ll.mit.edu/fuse/examples/simpleService/1.0\"") );

  }


  /** 
   *  Test of global edit vs first match only
   *
   */
  @Test
  public void globalEditTest()
    throws IOException, PatternSyntaxException, Exception {

    //
    // Create processor with single expression to capitalize all
    // instances of 'sed'
    //
    SedProcessor processor = new SedProcessor("s/sed/Sed/g" );

    _inMsg.setBody("sed is old, but sed is still useful");

    processor.process( _exchange );

    String outBody = _outMsg.getBody( String.class );

    assertTrue( outBody.equals("Sed is old, but Sed is still useful") ); 


    //
    // Create processor with single expression to capitalize first
    // instances of 'sed' in message
    //
    processor = new SedProcessor("s/sed/Sed/" );

    _inMsg.setBody("sed is old, but sed is still useful");

    processor.process( _exchange );

    outBody = _outMsg.getBody( String.class );

    assertTrue( outBody.equals("Sed is old, but sed is still useful") ); 

  }


  /** 
   *  Test of matches that ignore uppercase/lowercase differences
   */
  @Test
  public void ignoreCaseInMatchTest()
    throws IOException, PatternSyntaxException, Exception {

    //
    // Create processor with single expression to convert all
    // instances of lowercase 'sed' to ALLCAPS
    //
    SedProcessor processor = new SedProcessor("s/sed/SED/g");

    _inMsg.setBody("Sed is old, but sed is still useful");

    processor.process( _exchange );

    String outBody = _outMsg.getBody( String.class );

    assertTrue( outBody.equals("Sed is old, but SED is still useful") ); 


    //
    // Create processor with single expression to convert all
    // instances of 'sed' to ALLCAPS, independent of case ('i' option)
    //
    processor = new SedProcessor("s/sed/SED/gi");

    _inMsg.setBody("sed is old, but sed is still useful");

    processor.process( _exchange );

    outBody = _outMsg.getBody( String.class );

    assertTrue( outBody.equals("SED is old, but SED is still useful") ); 

  }


  /** 
   *  Test of the feature that allows for groups in the match pattern to be
   *  used in the replace pattern
   */
  @Test
  public void groupReplaceTest()
    throws IOException, PatternSyntaxException, Exception {


    //
    // Load file with SED expression. Rule in file is specified as: 
    //
    //  s/(\w+) is the next big thing/\1 is a legacy technology/g
    //
    // This expression modifies a string of the from:
    // 
    //   "X is the next big thing"
    // 
    //   to:
    // 
    //   "X is a legacy technology
    // 
    //  where X is any word. 

    SedProcessor processor = 
      new SedProcessor( new File("./src/test/resources/groupReplace") );


    _inMsg.setBody("SOAP is the next big thing");

    processor.process( _exchange );

    String outBody = _outMsg.getBody( String.class );

    //System.out.println("Message outMsg body: " + outBody );

    assertTrue( outBody.equals("SOAP is a legacy technology") );

  }



}
